@@ -1,5 +1,5 @@
 /* util.c -- functions for initializing new tree elements, and other things.
-   Copyright (C) 1990, 91, 92, 93, 94 Free Software Foundation, Inc.
+   Copyright (C) 1990, 91, 92, 93, 94, 2000 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -15,31 +15,20 @@
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */
 
-#include <config.h>
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <stdio.h>
 #include "defs.h"
-
-/* Return the last component of pathname FNAME, with leading slashes
-   compressed into one slash. */
 
-#ifndef __linux__
-char *
-basename (fname)
-     char *fname;
-{
-  char *p;
+#if ENABLE_NLS
+# include <libintl.h>
+# define _(Text) gettext (Text)
+#else
+# define _(Text) Text
+#endif
+#ifdef gettext_noop
+# define N_(String) gettext_noop (String)
+#else
+# define N_(String) (String)
+#endif
 
-  /* For "/", "//", etc., return "/". */
-  for (p = fname; *p == '/'; ++p)
-    /* Do nothing. */ ;
-  if (*p == '\0')
-    return p - 1;
-  p = strrchr (fname, '/');
-  return (p == NULL ? fname : p + 1);
-}
-#endif /* not __linux__ */
 
 /* Return a pointer to a new predicate structure, which has been
    linked in as the last one in the predicates list.
@@ -50,7 +39,7 @@ basename (fname)
    Set all cells in the new structure to the default values. */
 
 struct predicate *
-get_new_pred ()
+get_new_pred (void)
 {
   register struct predicate *new_pred;
 
@@ -86,7 +75,7 @@ get_new_pred ()
    predicate is an operator.  If it isn't, the AND operator is inserted. */
 
 struct predicate *
-get_new_pred_chk_op ()
+get_new_pred_chk_op (void)
 {
   struct predicate *new_pred;
 
@@ -94,7 +83,7 @@ get_new_pred_chk_op ()
     switch (last_pred->p_type)
       {
       case NO_TYPE:
-	error (1, 0, "oops -- invalid default insertion of and!");
+	error (1, 0, _("oops -- invalid default insertion of and!"));
 	break;
 
       case PRIMARY_TYPE:
@@ -132,8 +121,7 @@ get_new_pred_chk_op ()
    operator. */
 
 struct predicate *
-insert_primary (pred_func)
-     boolean (*pred_func) ();
+insert_primary (boolean (*pred_func) (/* ??? */))
 {
   struct predicate *new_pred;
 
@@ -149,12 +137,11 @@ insert_primary (pred_func)
 }
 
 void
-usage (msg)
-     char *msg;
+usage (char *msg)
 {
   if (msg)
     fprintf (stderr, "%s: %s\n", program_name, msg);
-  fprintf (stderr, "\
-Usage: %s [path...] [expression]\n", program_name);
+  fprintf (stderr, _("\
+Usage: %s [path...] [expression]\n"), program_name);
   exit (1);
 }
